from operator import attrgetter

import yaml

from toolrack.testing import (
    TestCase,
    TempDirFixture)

from ..config import load_config, ConfigError


class LoadConfigTests(TestCase):

    def setUp(self):
        super().setUp()
        self.tempdir = self.useFixture(TempDirFixture())
        self.config = {
            'databases': {'db': {'dsn': 'postgres:///foo'}},
            'metrics': {'m': {'type': 'gauge'}},
            'queries': {
                'q': {
                    'interval': 10,
                    'databases': ['db'],
                    'metrics': ['m'],
                    'sql': 'SELECT 1'}}}

    def test_load_databases_section(self):
        """The 'databases' section is loaded from the config file."""
        config = {
            'databases': {
                'db1': {'dsn': 'postgres:///foo'},
                'db2': {'dsn': 'postgres:///bar'}}}
        config_file = self.tempdir.mkfile(content=yaml.dump(config))
        with open(config_file) as fd:
            result = load_config(fd)
        database1, database2 = sorted(result.databases, key=attrgetter('name'))
        self.assertEqual(database1.name, 'db1')
        self.assertEqual(database1.dsn, 'postgres:///foo')
        self.assertEqual(database2.name, 'db2')
        self.assertEqual(database2.dsn, 'postgres:///bar')

    def test_load_databases_missing_dsn(self):
        """An error is raised if the 'dsn' key is missing for a database."""
        config = {'databases': {'db1': {}}}
        config_file = self.tempdir.mkfile(content=yaml.dump(config))
        with self.assertRaises(ConfigError) as cm, open(config_file) as fd:
            load_config(fd)
        self.assertEqual(
            str(cm.exception), "Missing key 'dsn' for database 'db1'")

    def test_load_metrics_section(self):
        """The 'metrics' section is loaded from the config file."""
        config = {
            'metrics': {
                'metric1': {
                    'type': 'summary',
                    'description': 'metric one'},
                'metric2': {
                    'type': 'histogram',
                    'description': 'metric two',
                    'buckets': [10, 100, 1000]}}}
        config_file = self.tempdir.mkfile(content=yaml.dump(config))
        with open(config_file) as fd:
            result = load_config(fd)
        metric1, metric2 = sorted(result.metrics, key=attrgetter('name'))
        self.assertEqual(metric1.type, 'summary')
        self.assertEqual(metric1.description, 'metric one')
        self.assertEqual(metric1.config, {'labels': ['database']})
        self.assertEqual(metric2.type, 'histogram')
        self.assertEqual(metric2.description, 'metric two')
        self.assertEqual(
            metric2.config,
            {'labels': ['database'], 'buckets': [10, 100, 1000]})

    def test_load_queries_section(self):
        """The 'queries section is loaded from the config file."""
        config = {
            'databases': {
                'db1': {'dsn': 'postgres:///foo'},
                'db2': {'dsn': 'postgres:///bar'}},
            'metrics': {
                'm1': {'type': 'summary'},
                'm2': {'type': 'histogram'}},
            'queries': {
                'q1': {
                    'interval': 10,
                    'databases': ['db1'],
                    'metrics': ['m1'],
                    'sql': 'SELECT 1'},
                'q2': {
                    'interval': 10,
                    'databases': ['db2'],
                    'metrics': ['m2'],
                    'sql': 'SELECT 2'}}}
        config_file = self.tempdir.mkfile(content=yaml.dump(config))
        with open(config_file) as fd:
            result = load_config(fd)
        query1, query2 = sorted(result.queries, key=attrgetter('name'))
        self.assertEqual(query1.name, 'q1')
        self.assertEqual(query1.databases, ['db1'])
        self.assertEqual(query1.metrics, ['m1'])
        self.assertEqual(query1.sql, 'SELECT 1')
        self.assertEqual(query2.name, 'q2')
        self.assertEqual(query2.databases, ['db2'])
        self.assertEqual(query2.metrics, ['m2'])
        self.assertEqual(query2.sql, 'SELECT 2')

    def test_load_queries_unknown_databases(self):
        """An error is raised if database names in query config are unknown."""
        config = {
            'metrics': {'m': {'type': 'summary'}},
            'queries': {
                'q': {
                    'interval': 10,
                    'databases': ['db1', 'db2'],
                    'metrics': ['m'],
                    'sql': 'SELECT 1'}}}
        config_file = self.tempdir.mkfile(content=yaml.dump(config))
        with self.assertRaises(ConfigError) as cm, open(config_file) as fd:
            load_config(fd)
        self.assertEqual(
            str(cm.exception), "Unknown databases for query 'q': db1, db2")

    def test_load_queries_unknown_metrics(self):
        """An error is raised if metric names in query config are unknown."""
        config = {
            'databases': {
                'db': {'dsn': 'postgres:///foo'}},
            'queries': {
                'q': {
                    'interval': 10,
                    'databases': ['db'],
                    'metrics': ['m1', 'm2'],
                    'sql': 'SELECT 1'}}}
        config_file = self.tempdir.mkfile(content=yaml.dump(config))
        with self.assertRaises(ConfigError) as cm, open(config_file) as fd:
            load_config(fd)
        self.assertEqual(
            str(cm.exception), "Unknown metrics for query 'q': m1, m2")

    def test_load_queries_missing_key(self):
        """An error is raised if keys are missing in the queries section."""
        config = {'queries': {'q1': {'interval': 10}}}
        config_file = self.tempdir.mkfile(content=yaml.dump(config))
        with self.assertRaises(ConfigError) as cm, open(config_file) as fd:
            load_config(fd)
        self.assertEqual(
            str(cm.exception), "Missing key 'databases' for query 'q1'")

    def test_load_queries_missing_interval_default_to_none(self):
        """If the interval is not specified, it defaults to None."""
        config = {
            'databases': {
                'db': {'dsn': 'postgres:///foo'}},
            'metrics': {
                'm': {'type': 'summary'}},
            'queries': {
                'q': {
                    'databases': ['db'],
                    'metrics': ['m'],
                    'sql': 'SELECT 1'}}}
        config_file = self.tempdir.mkfile(content=yaml.dump(config))
        with open(config_file) as fd:
            config = load_config(fd)
        self.assertIsNone(config.queries[0].interval)

    def test_load_queries_interval_seconds(self):
        """The query interval is stored as seconds."""
        config_file = self.tempdir.mkfile(content=yaml.dump(self.config))
        with open(config_file) as fd:
            config = load_config(fd)
        [query] = config.queries
        self.assertEqual(query.interval, 10)

    def test_load_queries_interval_as_string(self):
        """The interval can be passed as string."""
        self.config['queries']['q']['interval'] = '10'
        config_file = self.tempdir.mkfile(content=yaml.dump(self.config))
        with open(config_file) as fd:
            config = load_config(fd)
        [query] = config.queries
        self.assertEqual(query.interval, 10)

    def test_load_queries_interval_second_suffix(self):
        """The 's' suffix can be used in interval values for seconds."""
        self.config['queries']['q']['interval'] = '10s'
        config_file = self.tempdir.mkfile(content=yaml.dump(self.config))
        with open(config_file) as fd:
            config = load_config(fd)
        [query] = config.queries
        self.assertEqual(query.interval, 10)

    def test_load_queries_interval_minute_suffix(self):
        """The 'm' suffix can be used in interval values for minutes."""
        self.config['queries']['q']['interval'] = '10m'
        config_file = self.tempdir.mkfile(content=yaml.dump(self.config))
        with open(config_file) as fd:
            config = load_config(fd)
        [query] = config.queries
        self.assertEqual(query.interval, 600)

    def test_load_queries_interval_hour(self):
        """The 'h' suffix can be used in interval values for hours."""
        self.config['queries']['q']['interval'] = '1h'
        config_file = self.tempdir.mkfile(content=yaml.dump(self.config))
        with open(config_file) as fd:
            config = load_config(fd)
        [query] = config.queries
        self.assertEqual(query.interval, 3600)

    def test_load_queries_interval_day(self):
        """The 'd' suffix can be used in interval values for days."""
        self.config['queries']['q']['interval'] = '1d'
        config_file = self.tempdir.mkfile(content=yaml.dump(self.config))
        with open(config_file) as fd:
            config = load_config(fd)
        [query] = config.queries
        self.assertEqual(query.interval, 3600 * 24)

    def test_load_queries_interval_none(self):
        """The interval can be None."""
        self.config['queries']['q']['interval'] = None
        config_file = self.tempdir.mkfile(content=yaml.dump(self.config))
        with open(config_file) as fd:
            config = load_config(fd)
        [query] = config.queries
        self.assertIsNone(query.interval)

    def test_load_queries_interval_not_specified(self):
        """If the interval is not specified, it's set to None."""
        del self.config['queries']['q']['interval']
        config_file = self.tempdir.mkfile(content=yaml.dump(self.config))
        with open(config_file) as fd:
            config = load_config(fd)
        [query] = config.queries
        self.assertIsNone(query.interval)

    def test_invalid_interval_suffix(self):
        """An invalid suffix for query interval raises an error."""
        self.config['queries']['q']['interval'] = '1x'
        config_file = self.tempdir.mkfile(content=yaml.dump(self.config))
        with self.assertRaises(ConfigError) as cm, open(config_file) as fd:
            load_config(fd)
        self.assertEqual(str(cm.exception), "Invalid interval for query 'q'")

    def test_invalid_interval_not_number(self):
        """An query interval that is not a number raises an error."""
        self.config['queries']['q']['interval'] = 'wrong'
        config_file = self.tempdir.mkfile(content=yaml.dump(self.config))
        with self.assertRaises(ConfigError) as cm, open(config_file) as fd:
            load_config(fd)
        self.assertEqual(str(cm.exception), "Invalid interval for query 'q'")

    def test_invalid_interval_not_integer(self):
        """An query interval that is not an integer raises an error."""
        self.config['queries']['q']['interval'] = '1.5m'
        config_file = self.tempdir.mkfile(content=yaml.dump(self.config))
        with self.assertRaises(ConfigError) as cm, open(config_file) as fd:
            load_config(fd)
        self.assertEqual(str(cm.exception), "Invalid interval for query 'q'")

    def test_invalid_interval_zero(self):
        """An query interval of zero raises an error."""
        self.config['queries']['q']['interval'] = 0
        config_file = self.tempdir.mkfile(content=yaml.dump(self.config))
        with self.assertRaises(ConfigError) as cm, open(config_file) as fd:
            load_config(fd)
        self.assertEqual(str(cm.exception), "Invalid interval for query 'q'")

    def test_invalid_interval_negative(self):
        """An negative query interval raises an error."""
        self.config['queries']['q']['interval'] = -20
        config_file = self.tempdir.mkfile(content=yaml.dump(self.config))
        with self.assertRaises(ConfigError) as cm, open(config_file) as fd:
            load_config(fd)
        self.assertEqual(str(cm.exception), "Invalid interval for query 'q'")



import logging

import yaml

from fixtures import LoggerFixture

from prometheus_aioexporter.metric import MetricsRegistry

from toolrack.testing import TempDirFixture
from toolrack.testing.async import LoopTestCase

from .fakes import FakeSQLAlchemy
from ..config import load_config
from ..loop import QueryLoop


class QueryLoopTests(LoopTestCase):

    def setUp(self):
        super().setUp()
        self.tempdir = self.useFixture(TempDirFixture())
        self.logger = self.useFixture(LoggerFixture(level=logging.DEBUG))
        self.registry = MetricsRegistry()
        config_struct = {
            'databases': {'db': {'dsn': 'postgres:///foo'}},
            'metrics': {'m': {'type': 'gauge'}},
            'queries': {
                'q': {
                    'interval': 10,
                    'databases': ['db'],
                    'metrics': ['m'],
                    'sql': 'SELECT 1'},
                'q-no-interval': {
                    'databases': ['db'],
                    'metrics': ['m'],
                    'sql': 'SELECT 2'}
            }}
        config_file = self.tempdir.mkfile(content=yaml.dump(config_struct))
        with open(config_file) as fh:
            config = load_config(fh)

        self.registry.create_metrics(config.metrics)
        self.query_loop = QueryLoop(config, self.registry, logging, self.loop)
        self.query_loop._databases['db'].sqlalchemy = FakeSQLAlchemy()

    def mock_execute_query(self):
        """Don't actually execute queries."""
        self.query_exec = []

        async def _execute_query(*args):
            self.query_exec.append(args)

        self.query_loop._execute_query = _execute_query

    async def test_start(self):
        """The start method starts periodic calls for queries."""
        self.mock_execute_query()
        await self.query_loop.start()
        [periodic_call] = self.query_loop._periodic_calls
        self.assertTrue(periodic_call.running)
        await self.query_loop.stop()

    async def test_stop(self):
        """The stop method stops periodic calls for queries."""
        self.mock_execute_query()
        await self.query_loop.start()
        [periodic_call] = self.query_loop._periodic_calls
        await self.query_loop.stop()
        self.assertFalse(periodic_call.running)

    async def test_run_query(self):
        """Queries are run and update metrics."""
        database = self.query_loop._databases['db']
        database.sqlalchemy = FakeSQLAlchemy(query_results=[(100.0,)])
        await self.query_loop.start()
        await self.query_loop.stop()
        metric = self.registry.get_metric('m')
        # the metric is updated
        [(_, _, value)] = metric._samples()
        self.assertEqual(value, 100.0)

    async def test_run_query_null_value(self):
        """A null value in query results is treated like a zero."""
        database = self.query_loop._databases['db']
        database.sqlalchemy = FakeSQLAlchemy(query_results=[(None,)])
        await self.query_loop.start()
        await self.query_loop.stop()
        metric = self.registry.get_metric('m')
        [(_, _, value)] = metric._samples()
        self.assertEqual(value, 0)

    async def test_run_query_log(self):
        """Debug messages are logged on query execution."""
        database = self.query_loop._databases['db']
        database.sqlalchemy = FakeSQLAlchemy(query_results=[(100.0,)])
        await self.query_loop.start()
        await self.query_loop.stop()
        self.assertIn('running query "q" on database "db"', self.logger.output)
        self.assertIn('updating metric "m" set(100.0)', self.logger.output)

    async def test_run_query_log_error(self):
        """Query errors are logged."""
        database = self.query_loop._databases['db']
        database.sqlalchemy = FakeSQLAlchemy(connect_error='error')
        await self.query_loop.start()
        await self.query_loop.stop()
        self.assertIn(
            'query "q" on database "db" failed: error', self.logger.output)

    async def test_run_query_log_invalid_result_count(self):
        """An error is logged if result count doesn't match metrics count."""
        database = self.query_loop._databases['db']
        database.sqlalchemy = FakeSQLAlchemy(query_results=[(100.0, 200.0)])
        await self.query_loop.start()
        await self.query_loop.stop()
        self.assertIn(
            'query "q" on database "db" failed: Wrong result count from the'
            ' query',
            self.logger.output)

    async def test_run_query_periodically(self):
        """Queries are run at the specified time interval."""
        self.mock_execute_query()
        database = self.query_loop._databases['db']
        database.sqlalchemy = FakeSQLAlchemy(query_results=[(100.0,)])
        await self.query_loop.start()
        # the query has been run once
        self.assertEqual(len(self.query_exec), 1)
        self.loop.advance(5)
        # no more runs yet
        self.assertEqual(len(self.query_exec), 1)
        # now the query runs again
        self.loop.advance(5)
        self.assertEqual(len(self.query_exec), 2)
        await self.query_loop.stop()

    async def test_run_aperiodic_queries(self):
        """Queries with null interval can be run explicitly."""
        self.mock_execute_query()
        database = self.query_loop._databases['db']
        database.sqlalchemy = FakeSQLAlchemy(query_results=[(100.0,)])
        await self.query_loop.start()
        await self.query_loop.run_aperiodic_queries()
        self.assertEqual(len(self.query_exec), 2)
        await self.query_loop.run_aperiodic_queries()
        self.assertEqual(len(self.query_exec), 3)
        await self.query_loop.stop()



"""Configuration management functions."""

from collections import (
    defaultdict,
    namedtuple)

import yaml

from prometheus_aioexporter.metric import MetricConfig

from .db import Query, DataBase


class ConfigError(Exception):
    """Configuration is invalid."""


# Top-level configuration
Config = namedtuple('Config', ['databases', 'metrics', 'queries'])


def load_config(config_fd):
    """Load YaML config from file."""
    config = defaultdict(dict, yaml.load(config_fd))
    databases = _get_databases(config['databases'])
    metrics = _get_metrics(config['metrics'])
    database_names = frozenset(database.name for database in databases)
    metric_names = frozenset(metric.name for metric in metrics)
    queries = _get_queries(config['queries'], database_names, metric_names)
    return Config(databases, metrics, queries)


def _get_databases(configs):
    """Return a dict mapping names to DataBases names."""
    databases = []
    for name, config in configs.items():
        try:
            databases.append(DataBase(name, config['dsn']))
        except KeyError as e:
            _raise_missing_key(e, 'database', name)
    return databases


def _get_metrics(metrics):
    """Return metrics configuration."""
    configs = []
    for name, config in metrics.items():
        metric_type = config.pop('type', '')
        description = config.pop('description', '')
        # add a 'database' label to have different series for sharded databases
        config['labels'] = ['database']
        configs.append(MetricConfig(name, description, metric_type, config))
    return configs


def _get_queries(configs, database_names, metric_names):
    """Return a list of Queries from config."""
    queries = []
    for name, config in configs.items():
        try:
            _validate_query_config(name, config, database_names, metric_names)
            _convert_interval(name, config)
            queries.append(
                Query(
                    name, config['interval'], config['databases'],
                    config['metrics'], config['sql'].strip()))
        except KeyError as e:
            _raise_missing_key(e, 'query', name)
    return queries


def _validate_query_config(name, config, database_names, metric_names):
    """Validate a query configuration stanza."""
    unknown_databases = set(config['databases']) - database_names
    if unknown_databases:
        raise ConfigError(
            "Unknown databases for query '{}': {}".format(
                name, ', '.join(sorted(unknown_databases))))
    unknown_metrics = set(config['metrics']) - metric_names
    if unknown_metrics:
        raise ConfigError(
            "Unknown metrics for query '{}': {}".format(
                name, ', '.join(sorted(unknown_metrics))))


def _convert_interval(name, config):
    """Convert query intervals to seconds."""
    interval = config.setdefault('interval', None)
    if interval is None:
        # the query should be run at every request
        return

    multiplier = 1

    config_error = ConfigError(
        "Invalid interval for query '{}'".format(name))

    if isinstance(interval, str):
        # convert to seconds
        multipliers = {'s': 1, 'm': 60, 'h': 3600, 'd': 3600 * 24}
        suffix = interval[-1]
        if suffix in multipliers:
            interval = interval[:-1]
            multiplier = multipliers[suffix]

        if '.' in interval:
            raise config_error

    try:
        interval = int(interval)
    except ValueError:
        raise config_error

    if interval <= 0:
        raise config_error

    config['interval'] = interval * multiplier


def _raise_missing_key(key_error, entry_type, entry_name):
    raise ConfigError(
        'Missing key {} for {} \'{}\''.format(
            str(key_error), entry_type, entry_name))



"""Fakes for testing."""


class FakeSQLAlchemy:
    """Fake sqlalchemy module."""

    dsn = None
    strategy = None
    engine = None

    def __init__(self, connect_error=None, query_results=None,
                 query_error=None, missing_module=None):
        self.connect_error = connect_error
        self.query_results = query_results
        self.query_error = query_error
        self.missing_module = missing_module

    def create_engine(self, dsn, strategy=None):
        self.dsn = dsn
        self.strategy = strategy
        if self.missing_module is not None:
            raise ImportError(
                'Failed to import module "{}"'.format(self.missing_module),
                name=self.missing_module)
        self.engine = FakeEngine(
            dsn, query_results=self.query_results,
            query_error=self.query_error, connect_error=self.connect_error)
        return self.engine


class FakeEngine:
    """Fake database engine."""

    connection = None

    def __init__(self, dsn, query_results=None, query_error=None,
                 connect_error=None):
        self.dsn = dsn
        self.query_results = query_results
        self.query_error = query_error
        self.connect_error = connect_error

    async def connect(self):
        if self.connect_error:
            raise Exception(self.connect_error)
        self.connection = FakeConnection(
            query_results=self.query_results, query_error=self.query_error)
        return self.connection


class FakeConnection:
    """Fake connection."""

    closed = False
    sql = None

    def __init__(self, query_results=None, query_error=None):
        self.query_results = query_results
        self.query_error = query_error

    async def close(self):
        self.closed = True

    async def execute(self, sql):
        if self.query_error:
            raise Exception(self.query_error)
        self.sql = sql
        return FakeQueryResult(self.query_results)


class FakeQueryResult:
    """Fake query result."""

    def __init__(self, query_results):
        self.query_results = query_results

    async def fetchall(self):
        return self.query_results



from unittest import TestCase

from toolrack.testing.async import LoopTestCase

from .fakes import (
    FakeSQLAlchemy,
    FakeConnection)
from ..db import (
    Query,
    DataBase,
    DataBaseError,
    InvalidResultCount)


class QueryTests(TestCase):

    def test_instantiate(self):
        """A query can be instantiated with the specified arguments."""
        query = Query(
            'query', 20, ['db1', 'db2'], ['metric1', 'metric2'], 'SELECT 1')
        self.assertEqual(query.name, 'query')
        self.assertEqual(query.interval, 20)
        self.assertEqual(query.databases, ['db1', 'db2'])
        self.assertEqual(query.metrics, ['metric1', 'metric2'])
        self.assertEqual(query.sql, 'SELECT 1')

    def test_results(self):
        """The results method returns a dict mapping metrics to results."""
        query = Query('query', 20, ['db'], ['metric1', 'metric2'], 'SELECT 1')
        rows = [(11, 22), (33, 44)]
        self.assertEqual(
            query.results(rows), {'metric1': (11, 33), 'metric2': (22, 44)})

    def test_results_wrong_result_count(self):
        """An error is raised if the result column count is wrong."""
        query = Query('query', 20, ['db'], ['metric'], 'SELECT 1, 2')
        rows = [(1, 2)]
        with self.assertRaises(InvalidResultCount):
            query.results(rows)

    def test_results_empty(self):
        """No error is raised if the result set is empty"""
        query = Query('query', 20, ['db'], ['metric'], 'SELECT 1, 2')
        rows = []
        self.assertEqual(query.results(rows), {})


class DataBaseTests(LoopTestCase):

    def setUp(self):
        super().setUp()
        self.db = DataBase('db', 'postgres:///foo')
        self.db.sqlalchemy = FakeSQLAlchemy()

    def test_instantiate(self):
        """A DataBase can be instantiated with the specified arguments."""
        self.assertEqual(self.db.name, 'db')
        self.assertEqual(self.db.dsn, 'postgres:///foo')

    async def test_connect(self):
        """The connect connects to the database."""
        await self.db.connect()
        self.assertIsInstance(self.db._conn, FakeConnection)
        self.assertEqual(self.db.sqlalchemy.dsn, 'postgres:///foo')

    async def test_connect_missing_engine_module(self):
        """An error is raised if a module for the engine is missing."""
        self.db.sqlalchemy = FakeSQLAlchemy(missing_module='psycopg2')
        with self.assertRaises(DataBaseError) as cm:
            await self.db.connect()
        self.assertEqual(str(cm.exception), 'module "psycopg2" not found')

    async def test_connect_error(self):
        """A DataBaseError is raised if database connection fails."""
        self.db.sqlalchemy = FakeSQLAlchemy(connect_error='some error')
        with self.assertRaises(DataBaseError) as cm:
            await self.db.connect()
        self.assertEqual(str(cm.exception), 'some error')

    async def test_close(self):
        """The close method closes database connection."""
        await self.db.connect()
        connection = self.db._conn
        await self.db.close()
        self.assertTrue(connection.closed)
        self.assertIsNone(self.db._conn)

    async def test_execute(self):
        """The execute method executes a query."""
        query = Query('query', 20, ['db'], ['metric1', 'metric2'], 'SELECT 1')
        sqlalchemy = FakeSQLAlchemy(query_results=[(10, 20), (30, 40)])
        self.db.sqlalchemy = sqlalchemy

        await self.db.connect()
        result = await self.db.execute(query)
        self.assertEqual(result, {'metric1': (10, 30), 'metric2': (20, 40)})
        self.assertEqual(sqlalchemy.engine.connection.sql, 'SELECT 1')

    async def test_execute_query_error(self):
        """If the query fails an error is raised."""
        query = Query('query', 20, ['db'], ['metric'], 'WRONG')
        sqlalchemy = FakeSQLAlchemy(query_error='wrong query')
        self.db.sqlalchemy = sqlalchemy

        await self.db.connect()
        with self.assertRaises(DataBaseError) as cm:
            await self.db.execute(query)
        self.assertEqual(str(cm.exception), 'wrong query')

    async def test_execute_not_connected(self):
        """The execute recconnects to the database if not connected."""
        query = Query('query', 20, ['db'], ['metric1', 'metric2'], 'SELECT 1')
        sqlalchemy = FakeSQLAlchemy(query_results=[(10, 20), (30, 40)])
        self.db.sqlalchemy = sqlalchemy

        result = await self.db.execute(query)
        self.assertEqual(result, {'metric1': (10, 30), 'metric2': (20, 40)})
        self.assertEqual(sqlalchemy.engine.connection.sql, 'SELECT 1')
        # the connection is kept for reuse
        self.assertFalse(sqlalchemy.engine.connection.closed)



"""Loop to periodically execute queries."""

import asyncio

from toolrack.async import PeriodicCall

from .db import DataBaseError


class QueryLoop:
    """Periodically performs queries."""

    _METRIC_METHODS = {
        'counter': 'inc',
        'gauge': 'set',
        'histogram': 'observe',
        'summary': 'observe'}

    def __init__(self, config, registry, logger, loop):
        self.loop = loop
        self._logger = logger
        self._registry = registry
        self._periodic_queries = []
        self._aperiodic_queries = []
        self._periodic_calls = []
        self._setup(config)

    async def start(self):
        """Start periodic queries execution."""
        for db in self._databases.values():
            try:
                await db.connect()
            except DataBaseError as error:
                self._log_db_error(db.name, error)
            else:
                self._logger.debug(
                    'connected to database "{}"'.format(db.name))

        for query in self._periodic_queries:
            call = PeriodicCall(self.loop, self._run_query, query)
            self._periodic_calls.append(call)
            call.start(query.interval)

    async def stop(self):
        """Stop periodic query execution."""
        coros = (call.stop() for call in self._periodic_calls)
        await asyncio.gather(*coros, loop=self.loop)
        self._periodic_calls = []
        coros = (db.close() for db in self._databases.values())
        await asyncio.gather(*coros, loop=self.loop)

    async def run_aperiodic_queries(self):
        coros = (
            self._execute_query(query, dbname)
            for query in self._aperiodic_queries
            for dbname in query.databases)
        await asyncio.gather(*coros, loop=self.loop)

    def _setup(self, config):
        """Initialize instance attributes."""
        self._metric_configs = {
            metric_config.name: metric_config
            for metric_config in config.metrics}
        self._databases = {
            database.name: database for database in config.databases}

        for query in config.queries:
            if query.interval is None:
                self._aperiodic_queries.append(query)
            else:
                self._periodic_queries.append(query)

    def _run_query(self, query):
        """Periodic task to run a query."""
        for dbname in query.databases:
            self.loop.create_task(self._execute_query(query, dbname))

    async def _execute_query(self, query, dbname):
        """'Execute a Query on a DataBase."""
        self._logger.debug(
            'running query "{}" on database "{}"'.format(query.name, dbname))
        try:
            results = await self._databases[dbname].execute(query)
        except DataBaseError as error:
            self._log_query_error(query.name, dbname, error)
            return

        for name, values in results.items():
            for value in values:
                self._update_metric(name, value, dbname)

    def _log_query_error(self, name, dbname, error):
        """Log an error related to database query."""
        prefix = 'query "{}" on database "{}" failed:'.format(name, dbname)
        self._logger.error('{} {}'.format(prefix, error))

    def _log_db_error(self, name, error):
        """Log a failed database query."""
        prefix = 'error from database "{}":'.format(name)
        self._logger.error('{} {}'.format(prefix, error))

    def _update_metric(self, name, value, dbname):
        """Update value for a metric."""
        if value is None:
            # don't fail is queries that count return NULL
            value = 0.0
        method = self._METRIC_METHODS[self._metric_configs[name].type]
        self._logger.debug(
            'updating metric "{}" {}({})'.format(name, method, value))
        metric = self._registry.get_metric(name, labels={'database': dbname})
        getattr(metric, method)(value)



"""Script entry point."""

import argparse

from toolrack.script import ErrorExitMessage

from prometheus_aioexporter.script import PrometheusExporterScript
from prometheus_aioexporter.metric import InvalidMetricType

from .config import load_config, ConfigError
from .loop import QueryLoop


class QueryExporterScript(PrometheusExporterScript):
    """Periodically run database queries and export results to Prometheus."""

    name = 'query-exporter'

    description = __doc__

    def configure_argument_parser(self, parser):
        parser.add_argument(
            'config', type=argparse.FileType('r'),
            help='configuration file')

    def configure(self, args):
        config = self._load_config(args.config)
        self.create_metrics(config.metrics)
        self.query_loop = QueryLoop(
            config, self.registry, self.logger, self.loop)

    async def on_application_startup(self, application):
        application['exporter'].set_metric_update_handler(
            self._update_handler)
        await self.query_loop.start()

    async def on_application_shutdown(self, application):
        await self.query_loop.stop()

    async def _update_handler(self, metrics):
        """Run queries with no specified interval on each request."""
        await self.query_loop.run_aperiodic_queries()

    def _load_config(self, config_file):
        """Load the application configuration."""
        try:
            config = load_config(config_file)
        except (InvalidMetricType, ConfigError) as error:
            raise ErrorExitMessage(str(error))
        finally:
            config_file.close()
        return config


script = QueryExporterScript()



from setuptools import setup, find_packages

from query_exporter import __version__, __doc__ as description

tests_require = ['asynctest']

config = {
    'name': 'query-exporter',
    'version': __version__,
    'license': 'GPLv3+',
    'description': description,
    'long_description': open('README.rst').read(),
    'author': 'Alberto Donato',
    'author_email': 'alberto.donato@gmail.com',
    'maintainer': 'Alberto Donato',
    'maintainer_email': 'alberto.donato@gmail.com',
    'url': 'https://github.com/albertodonato/query-exporter',
    'packages': find_packages(),
    'include_package_data': True,
    'entry_points': {'console_scripts': [
        'query-exporter = query_exporter.main:script']},
    'test_suite': 'query_exporter',
    'install_requires': [
        'aiohttp',
        'prometheus-client',
        'prometheus-aioexporter >= 1.4.0',
        'PyYaml',
        'SQLAlchemy',
        'sqlalchemy_aio',
        'toolrack'],
    'tests_require': tests_require,
    'extras_require': {'testing': tests_require},
    'keywords': 'sql metric prometheus exporter',
    'classifiers': [
        'Development Status :: 4 - Beta',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Utilities']}

setup(**config)



"""Database wrapper."""

from collections import namedtuple

import sqlalchemy

from sqlalchemy_aio import ASYNCIO_STRATEGY


class DataBaseError(Exception):
    """A database error."""


class InvalidResultCount(Exception):
    """Number of results from a query don't match metrics count."""

    def __init__(self):
        super().__init__('Wrong result count from the query')


Query = namedtuple(
    'Query', ['name', 'interval', 'databases', 'metrics', 'sql'])


class Query(Query):

    def results(self, records):
        """Return a dict with a tuple of values for each metric."""
        if not records:
            return {}

        if len(self.metrics) != len(records[0]):
            raise InvalidResultCount()
        return dict(zip(self.metrics, zip(*records)))


class DataBase(namedtuple('DataBase', ['name', 'dsn'])):
    """A database to perform Queries."""

    sqlalchemy = sqlalchemy  # for testing

    _conn = None

    async def connect(self):
        """Connect to the database."""
        try:
            engine = self.sqlalchemy.create_engine(
                self.dsn, strategy=ASYNCIO_STRATEGY)
        except ImportError as error:
            raise DataBaseError('module "{}" not found'.format(error.name))

        try:
            self._conn = await engine.connect()
        except Exception as error:
            raise DataBaseError(str(error).strip())

    async def close(self):
        """Close the database connection."""
        if self._conn is None:
            return
        await self._conn.close()
        self._conn = None

    async def execute(self, query):
        """Execute a query."""
        if self._conn is None:
            await self.connect()

        try:
            result = await self._conn.execute(query.sql)
            return query.results(await result.fetchall())
        except Exception as error:
            raise DataBaseError(str(error).strip())
